/*-
 * Copyright (C) @BABOLO  2008 Apr 14
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 Apr 14\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: db_0.c,v 1.14 2009/07/14 05:05:29 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define INDX(R, C) [R * res->ncol + C]  /*        */

                                                                          /* Pc  db_0  NULL */

typedef struct pgoblin_db_0_data {                                                   /*  Pq */
    char    **cell;                                 /*       */
    size_t   *llen;      /*                          *
                          *  ,      strlen() */
    u_int32_t flags;
#   define    PGOBLIN_DB_0FREEC 0x0001                          /*     free() */
#   define    PGOBLIN_DB_0FREEA 0x0002                                           /* cell  free() */
    int       nrow;                                                    /*     */
    int       ncol;                                                  /*     */
} pgoblin_db_0_data;

static int
pgoblin_finit(u_int32_t flags, pgoblin_dbases *this) {
    return(0);
}

static void *
pgoblin_no() {
    warnx("Unimplemented in 0");
    return(NULL);
}

void *  /* Public procedure, used in pGoblin internally */
pgoblin_db_0_ctl(pgoblin_exenv *exenv, pgoblin_r *r) {
    pgoblin_db_0_data *db0;
    int c, i;

    ifBLIN_QV5(exenv->options->flags) fprintf( stderr, "+pgoblin_db_0_ctl\n");
    if  (!(db0 = malloc(sizeof(pgoblin_db_0_data)))) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_db_0_ctl: No mem");
        errno = ENOMEM;
        goto out;
    }
    bzero(db0, sizeof(pgoblin_db_0_data));
    db0->flags = exenv->options->flags & BLIN_VERMASK;
    db0->nrow = 1;
    for (c = 0; ; c++) {
        i = c;
        if  (c && exenv->pgm->code[exenv->ppoint + c].cmd != PGOBLIN_cont) break;
        if  (!exenv->pgm->code[exenv->ppoint + c].e.lit) break;
    }
    db0->ncol = i;
    if  (i) {
        if  (!(db0->cell = malloc((i + 1) * sizeof(char**)))) {
            free(db0);
            errno = ENOMEM;
            goto out;
        }
        db0->flags |= PGOBLIN_DB_0FREEA;
        for (c = 0; c < i; c++) {
            db0->cell[c] = exenv->pgm->code[exenv->ppoint + c].e.lit;
            ifBLIN_QV6(db0->flags) fprintf( stderr, "[%d]%s%s\n"
                                          , c, db0->cell[c] ? "=" : ""
                                          , db0->cell[c] ? db0->cell[c] : ""
                                          );
        }
        db0->cell[i] = NULL;
        ifBLIN_QV6(db0->flags) fprintf(stderr, "[%d]\n", c);
    }
out:
    ifBLIN_QV5(exenv->options->flags)
        fprintf( stderr, "-pgoblin_db_0_ctl %" BLIN_X " %08X r=%d c=%d %s %s\n"
               , BLIN_I(db0), db0->flags, db0->nrow, db0->ncol
               , db0->cell ? "cell" : "", db0->llen ? "llen" : ""
               );
    return(db0);
}

static int
pgoblin_shurecon(pgoblin_conn *db) {
    int ex = EX_OK;

    if  (!db) {
        ex = EX_DATAERR;
        errno = EINVAL;
    } else {
        /** R_CONN + **/
        ;;  db->conn = NULL;                                                              /* Pc  */
        /** R_CONN - **/
    }
    ifBLIN_QV5(db->flags) fprintf( stderr, "local pgoblin_shurecon %d\n", ex);
    return(ex);
}

static int  /*  #select      NULL,    query "rows cols" */
pgoblin_query(pgoblin_conn *db, void **res, const char *query, u_char cond) {
    pgoblin_db_0_data *db0;
    int ex = EX_OK;
    char *q;

    if  (!db) {
        errno = EFAULT;
        ex = EX_DATAERR;
    } else if (res) {
        if  (!(db0 = malloc(sizeof(pgoblin_db_0_data)))) {
            ifBLIN_QV1(*(u_int32_t*)(db->conn)) warnx("pgoblin_db_0_ctl: No mem");
            errno = ENOMEM;
            ex = EX_OSERR;
        } else {
            bzero(db0, sizeof(pgoblin_db_0_data));
            db0->flags = db->flags & BLIN_VERMASK;
            if  (query) {
                db0->nrow = strtoul(query, &q, 0);
                db0->ncol = strtoul(q, &q, 0);
            }
            *res = db0;
    }   }
    ifBLIN_QV5(db->flags) fprintf( stderr, "local pgoblin_query %d\n", ex);
    return(ex);
}

static int
pgoblin_ntuples(pgoblin_db_0_data *res) {
    int ex;

    if  (!res) {
        errno = EFAULT;
        ex = -1;
    } else {
        ex = res->nrow;
    }
    ifBLIN_QV5(res->flags) fprintf(stderr, "local pgoblin_ntuples %d\n", ex);
    return(ex);
}

static int
pgoblin_nfields(pgoblin_db_0_data *res) {
    int ex;

    if  (!res) {
        errno = EFAULT;
        ex = -1;
    } else {
        ex = res->ncol;
    }
    ifBLIN_QV5(res->flags) fprintf(stderr, "local pgoblin_nfields %d\n", ex);
    return(ex);
}

static int
pgoblin_isnull(pgoblin_db_0_data *res, int row, int col) {
    int e;

    if  (!res) {
        warnx("No res in local pgoblin_isnull\n");
        e = (row || col) ? -1 : 0;
    } else if (!res->cell) {
        e = 1;
    } else if (row >= 0 && row < res->nrow && col >= 0 && col < res->ncol) {
        e = !res->cell INDX(row, col);
    } else {
        ifBLIN_QV2(res->flags) {
            warnx( "db 0: row %d%s in [0..%d], col %d%s in [0..%d]"
                 , row, (row >= 0 && row < res->nrow) ? "" : " NOT", res->nrow
                 , col, (col >= 0 && col < res->ncol) ? "" : " NOT", res->ncol
                 );
        }
        e = -1;
    }
    ifBLIN_QV5(res->flags) fprintf(stderr, "local pgoblin_isnull(%d,%d) %d\n", row, col, e);
    return(e);
}

static int
pgoblin_length(pgoblin_db_0_data *res, int row, int col) {
    int e;

    if  ((e = pgoblin_isnull(res, row, col)) < 0) {
    } else if (e) {
        e = 0;
    } else if (res->llen) {
        e = res->llen INDX(row, col);
    } else if (res->cell) {
        e = strlen(res->cell INDX(row, col));
    }
    ifBLIN_QV5(res->flags) fprintf(stderr, "local pgoblin_length(%d,%d) %d\n", row, col, e);
    return(e);
}

static char *
pgoblin_value(pgoblin_db_0_data *res, int row, int col) {
    char *c = NULL;
    if  (!pgoblin_isnull(res, row, col)) c = res->cell INDX(row, col);
    ifBLIN_QV5(res->flags) fprintf(stderr, "local pgoblin_value(%d,%d)=%s~\n", row, col, c);
    return(c);
}

static void
pgoblin_frees(pgoblin_db_0_data **res) {
    int i;

    ifBLIN_QV5((*res)->flags) fprintf(stderr, "local pgoblin_frees\n");
    if  (res && *res) {
        if  ((*res)->llen) {
            free((*res)->llen);
            (*res)->llen = NULL;
        }
        if  ((*res)->cell) {
            if  ((*res)->flags & PGOBLIN_DB_0FREEC) {
                for (i = 0; i < (*res)->nrow * (*res)->ncol; i++) {
                    free((*res)->cell[i]);
                    (*res)->cell[i] = NULL;
            }   }
            if  ((*res)->flags & PGOBLIN_DB_0FREEA) {
                free((*res)->cell);
                (*res)->cell = NULL;
        }   }
        free(*res);
        *res = NULL;
}   }

static void
pgoblin_none(void *v) {}

static void
pgoblin_finish(void **conn) {
    if  (conn) *conn = NULL;
}

static int
pgoblin_tarry(void *conn, int timeout) {
    return(0);
}

pgoblin_dbases pgoblin_db_0 =
{ "3pGoblin-" VERS
, "0\0" VERS
,                                      pgoblin_finit                               /* init      */
,                                      pgoblin_finit                               /* fini      */
,                                      pgoblin_shurecon                            /* shurecon  */
, (int  (*)(pgoblin_conn*, void**, const char*, u_char))pgoblin_query              /* exec      */
, (int  (*)(void*))                    pgoblin_ntuples                             /* ntuples   */
, (int  (*)(void*))                    pgoblin_nfields                             /* nfields   */
, (int  (*)(void*, int, int))          pgoblin_isnull                              /* getisnull */
, (int  (*)(void*, int, int))          pgoblin_length                              /* getlength */
, (char*(*)(void*, int, int))          pgoblin_value                               /* getvalue  */
, (int  (*)(pgoblin_conn*, void**, int, u_char))pgoblin_no                         /* getstream */
, (int  (*)(pgoblin_conn*, void**))    pgoblin_no                                  /* getcopy   */
, (int  (*)(pgoblin_conn*, const void*, int))pgoblin_no                            /* putcopy   */
, (int  (*)(pgoblin_conn*, u_char))    pgoblin_no                                  /* endstream */
, (int  (*)(void*, int))               pgoblin_tarry                               /* tarry     */
, (char*(*)(void*))                    pgoblin_none                                /* notifies  */
, (void (*)(void**))                   pgoblin_none                                /* freecopy  */
, (void (*)(char**))                   pgoblin_none                                /* freenote  */
, (void (*)(void**))                   pgoblin_frees                               /* clear     */
, (void (*)(void**))                   pgoblin_finish                              /* finish    */
};
